import torch
import torch.nn as nn
import matplotlib.pyplot as plt
import copy
import math

from scipy.spatial import KDTree
from scipy.sparse.csgraph import dijkstra


from LAMINAR.utils.gaussian2uniform import gaussian_to_sphere, jacobian_gaussian_to_sphere

from torch.func import vmap


# define activation functions
def antiderivTanh(x):
    return torch.abs(x) + torch.log(1 + torch.exp(-2.0 * torch.abs(x)))

def derivTanh(x):
    return 1 - torch.pow(torch.tanh(x), 2)


# define a ResNet module

class ResNet(nn.Module):
    def __init__(self, d: int, m: int, nTh: int = 2):
        """
            ResNet N portion of Phi
        :param d:   int, dimension of space input (expect inputs to be d+1 for space-time)
        :param m:   int, hidden dimension
        :param nTh: int, number of resNet layers , (number of theta layers)
        """
        super().__init__()
        # catch excepions
        if nTh < 2:
            print("nTh must be an integer >= 2")
            exit(1)

        self.d = d
        self.m = m
        self.nTh = nTh

        self.layers = nn.ModuleList()
        self.layers.append(nn.Linear(d + 1, m, bias=True)) # opening layer
        self.layers.append(nn.Linear(m,m, bias=True)) # resnet layers
        for i in range(nTh-2):
            self.layers.append(copy.deepcopy(self.layers[1]))
        self.act = antiderivTanh
        self.h = 1.0 / (self.nTh-1) # step size for the ResNet


    def forward(self, x):
        """
            N(s;theta). the forward propogation of the ResNet
        :param x: tensor nex-by-d+1, inputs (+1 is time?)
        :return:  tensor nex-by-m,   outputs
        """

        x = self.act(self.layers[0].forward(x))

        for i in range(1, self.nTh):
            x = x + self.h * self.act(self.layers[i].forward(x))    # this is the resnet structure

        return x
    

def odefun(x, t, net, alph=[1.0,1.0,1.0]):
    """
    neural ODE combining the characteristics and log-determinant (see Eq. (2)), the transport costs (see Eq. (5)), and
    the HJB regularizer (see Eq. (7)).

    d_t  [x ; l ; v ; r] = odefun( [x ; l ; v ; r] , t )

    x - particle position
    l - log determinant
    v - accumulated transport costs (Lagrangian)
    r - accumulates violation of HJB condition along trajectory
    """
    nex, d_extra = x.shape
    d = d_extra - 3

    z = nn.functional.pad(x[:, :d], (0, 1, 0, 0), value=t) # concatenate with the time t

    gradPhi, trH = net.trHess(z)

    dx = -(1.0/alph[0]) * gradPhi[:,0:d]
    dl = -(1.0/alph[0]) * trH.unsqueeze(1)
    dv = 0.5 * torch.sum(torch.pow(dx, 2) , 1 ,keepdims=True)
    dr = torch.abs(-gradPhi[:,-1].unsqueeze(1) + alph[0] * dv) 
    
    return torch.cat((dx,dl,dv,dr), 1).to(x.device)


def stepRK4(odefun, z, Phi, alph, t0, t1):
    """
        Runge-Kutta 4 integration scheme
    :param odefun: function to apply at every time step
    :param z:      tensor nex-by-d+4, inputs
    :param Phi:    Module, the Phi potential function
    :param alph:   list, the 3 alpha values for the OT-Flow Problem
    :param t0:     float, starting time
    :param t1:     float, end time
    :return: tensor nex-by-d+4, features at time t1
    """

    h = t1 - t0 # step size
    z0 = z

    K = h * odefun(z0, t0, Phi, alph=alph)
    z = z0 + (1.0/6.0) * K
    
    K = h * odefun( z0 + 0.5*K , t0+(h/2) , Phi, alph=alph)
    #z += (2.0/6.0) * K
    z = z + (2.0/6.0) * K

    K = h * odefun( z0 + 0.5*K , t0+(h/2) , Phi, alph=alph)
    #z += (2.0/6.0) * K
    z = z + (2.0/6.0) * K

    K = h * odefun( z0 + K , t0+h , Phi, alph=alph)
    #z += (1.0/6.0) * K
    z = z + (1.0/6.0) * K

    return z

def stepRK1(odefun, z, Phi, alph, t0, t1):
    """
        Runge-Kutta 1 / Forward Euler integration scheme.  Added for comparison, but we recommend stepRK4.
    :param odefun: function to apply at every time step
    :param z:      tensor nex-by-d+4, inputs
    :param Phi:    Module, the Phi potential function
    :param alph:   list, the 3 alpha values for the mean field game problem
    :param t0:     float, starting time
    :param t1:     float, end time
    :return: tensor nex-by-d+4, features at time t1
    """
    #z += (t1 - t0) * odefun(z, t0, Phi, alph=alph)
    z = z + (t1 - t0) * odefun(z, t0, Phi, alph=alph)
    return z


def integrate(x, net, tspan , nt, stepper="rk4", alph =[1.0,1.0,1.0], intermediates=False):
    """
        perform the time integration in the d-dimensional space
    :param x:       input data tensor nex-by-d
    :param net:     neural network Phi
    :param tspan:   time range to integrate over, ex. [0.0 , 1.0]
    :param nt:      number of time steps
    :param stepper: string "rk1" or "rk4" Runge-Kutta schemes
    :param alph:    list of length 3, the alpha value multipliers
    :param intermediates: bool, True means save all intermediate time points along trajectories
    :return:
        z - tensor nex-by-d+4, features at time t1
        OR zFull - tensor nex-by-d+3-by-nt+1 , trajectories from time t0 to t1 (when intermediates=True)
    """
    device = x.device
    h = (tspan[1]-tspan[0]) / nt

    # initialize "hidden" vector to propagate with all the additional dimensions for all the ODEs
    z = nn.functional.pad(x, (0, 3, 0, 0), value=tspan[0]).to(device)

    tk = tspan[0]

    if intermediates: # save the intermediate values as well
        #zFull = torch.zeros(*z.shape, nt+1, device=device, dtype=x.dtype) # make tensor of size z.shape[0], z.shape[1], nt
        #zFull[:,:,0] = z
        #zFull = torch.cat((z.unsqueeze(-1), zFull[:,:,1:]), dim=-1)
        with torch.no_grad():
            zFull = z.clone().reshape(z.shape[0], z.shape[1], 1)
            if stepper == 'rk4':
                for k in range(nt):
                    #zFull[:,:,k+1] = stepRK4(odefun, zFull[:,:,k], net, alph, tk, tk+h)
                    zFull = torch.cat((zFull, stepRK4(odefun, zFull[:,:,k], net, alph, tk, tk+h).unsqueeze(-1)), dim=-1)
                    tk += h
            elif stepper == 'rk1':
                for k in range(nt):
                    #zFull[:,:,k+1] = stepRK1(odefun, zFull[:,:,k], net, alph, tk, tk+h)
                    zFull = torch.cat((zFull, stepRK4(odefun, zFull[:,:,k], net, alph, tk, tk+h).unsqueeze(-1)), dim=-1)
                    tk += h

            return zFull

    else:
        with torch.no_grad():
            if stepper == 'rk4':
                for k in range(nt):
                    z = stepRK4(odefun, z, net, alph, tk, tk+h)
                    tk += h
            elif stepper == 'rk1':
                for k in range(nt):
                    z = stepRK1(odefun, z, net, alph, tk, tk+h)
                    tk += h

            return z

    # return in case of error
    return -1


class Phi(nn.Module):
    def __init__(self, nTh, m, d, r=10, alph=[1.0] * 5, device=torch.device('cpu')):
        """
            neural network approximating Phi (see Eq. (10) in OT paper)

            Phi( x,t ) = w'*ResNet( [x;t]) + 0.5*[x' t] * A'A * [x;t] + b'*[x;t] + c
            A, b, c are parameterising quadratic mechanics, so N is focussing on higher orders

        :param nTh:  int, number of resNet layers , (number of theta layers)
        :param m:    int, hidden dimension
        :param d:    int, dimension of space input (expect inputs to be d+1 for space-time)
        :param r:    int, rank r for the A matrix
        :param alph: list, alpha values / weighted multipliers for the optimization problem
        """

        self.device = device
        super().__init__()

        self.m    = m
        self.nTh  = nTh
        self.d    = d
        self.alph = alph

        r = min(r,d+1)

        self.A  = nn.Parameter(torch.zeros(r, d+1, device=device) , requires_grad=True)
        self.A  = nn.init.xavier_uniform_(self.A)
        self.c  = nn.Linear( d+1  , 1  , bias=True)  # b'*[x;t] + c (bias)
        self.w  = nn.Linear( m    , 1  , bias=False).to(device)

        self.N = ResNet(d, m, nTh).to(device)

        # set initial values
        self.w.weight.data = torch.ones(self.w.weight.data.shape, device=device)   # w = 1
        self.c.weight.data = torch.zeros(self.c.weight.data.shape, device=device)  # b = 0
        self.c.bias.data   = torch.zeros(self.c.bias.data.shape, device=device)    # c = 0

    def switch_device(self, device):
        self.device = device
        self.A = nn.Parameter(self.A.to(device))
        self.c.to(device)
        self.w.to(device)
        self.N.to(device)

    def forward(self, x):
        """ calculating Phi(s, theta)...not used in OT-Flow """

        # force A to be symmetric
        symA = torch.matmul(torch.t(self.A), self.A) # A'A

        return self.w( self.N(x)) + 0.5 * torch.sum( torch.matmul(x , symA) * x , dim=1, keepdims=True) + self.c(x)
    
    def trHess(self, x, justGrad=False):
        """
        compute gradient of Phi wrt x and trace(Hessian of Phi); see Eq. (11) and Eq. (13), respectively
        recomputes the forward propogation portions of Phi

        :param x: input data, torch Tensor nex-by-d
        :param justGrad: boolean, if True only return gradient, if False return (grad, trHess)
        :return: gradient , trace(hessian)    OR    just gradient
        """

        # code in E = eye(d+1,d) as index slicing instead of matrix multiplication
        # assumes specific N.act as the antiderivative of tanh

        N    = self.N
        m    = N.layers[0].weight.shape[0]
        nex  = x.shape[0] # number of examples in the batch
        d    = x.shape[1]-1
        symA = torch.matmul(self.A.t(), self.A)

        u = [] # hold the u_0,u_1,...,u_M for the forward pass
        z = N.nTh*[None] # hold the z_0,z_1,...,z_M for the backward pass
        # preallocate z because we will store in the backward pass and we want the indices to match the paper

        # Forward of ResNet N and fill u
        opening = N.layers[0].forward(x) # K_0 * S + b_0
        u.append(N.act(opening)) # u0
        feat = u[0]

        for i in range(1,N.nTh):
            feat = feat + N.h * N.act(N.layers[i](feat))
            u.append(feat)

        # going to be used more than once
        tanhopen = torch.tanh(opening) # act'( K_0 * S + b_0 )

        # compute gradient and fill z
        for i in range(N.nTh-1,0,-1): # work backwards, placing z_i in appropriate spot
            if i == N.nTh-1:
                term = self.w.weight.t()
            else:
                term = z[i+1]

            # z_i = z_{i+1} + h K_i' diag(...) z_{i+1}
            z[i] = term + N.h * torch.mm( N.layers[i].weight.t() , torch.tanh( N.layers[i].forward(u[i-1]) ).t() * term)

        # z_0 = K_0' diag(...) z_1
        z[0] = torch.mm( N.layers[0].weight.t() , tanhopen.t() * z[1] )
        grad = z[0] + torch.mm(symA, x.t() ) + self.c.weight.t()

        if justGrad:
            return grad.t()
        
        # -----------------
        # trace of Hessian
        #-----------------

        # t_0, the trace of the opening layer
        Kopen = N.layers[0].weight[:,0:d]    # indexed version of Kopen = torch.mm( N.layers[0].weight, E  )
        temp  = derivTanh(opening.t()) * z[1]
        trH  = torch.sum(temp.reshape(m, -1, nex) * torch.pow(Kopen.unsqueeze(2), 2), dim=(0, 1)) # trH = t_0

        # grad_s u_0 ^ T
        temp = tanhopen.t()   # act'( K_0 * S + b_0 )
        Jac  = Kopen.unsqueeze(2) * temp.unsqueeze(1) # K_0' * act'( K_0 * S + b_0 )
        # Jac is shape m by d by nex

        # t_i, trace of the resNet layers
        # KJ is the K_i^T * grad_s u_{i-1}^T
        for i in range(1,N.nTh):
            KJ  = torch.mm(N.layers[i].weight , Jac.reshape(m,-1) )
            KJ  = KJ.reshape(m,-1,nex)
            if i == N.nTh-1:
                term = self.w.weight.t()
            else:
                term = z[i+1]

            temp = N.layers[i].forward(u[i-1]).t() # (K_i * u_{i-1} + b_i)
            t_i = torch.sum(  ( derivTanh(temp) * term ).reshape(m,-1,nex)  *  torch.pow(KJ,2) ,  dim=(0, 1) )
            trH  = trH + N.h * t_i  # add t_i to the accumulate trace
            Jac = Jac + N.h * torch.tanh(temp).reshape(m, -1, nex) * KJ # update Jacobian

        return grad.t(), trH + torch.trace(symA[0:d,0:d])
        # indexed version of: return grad.t() ,  trH + torch.trace( torch.mm( E.t() , torch.mm(  symA , E) ) )
    
    def fullHessian(self, x, steps=1):
        # Calculate Hessian of Phi for different time steps specified by steps
        n_steps = torch.linspace(0, 1, steps + 1).reshape(-1, 1)[:-1].to(self.device)
    
        # Integrate over time in steps
        #alph = [1.0, 100.0, 5.0] # not needed here?
        #print(x.shape)
        zFull = integrate(x, self, [0, 1], nt=steps, stepper="rk4", intermediates=True)[:, :self.d, :-1].to(self.device)
        #zFull = torch.randn(x.shape[0], x.shape[1], steps, device=self.device)

        # Prepare data for batch processing
        dim = x.shape[1]
        batch_size = x.shape[0]
        time_dim = n_steps.shape[0]
    
        # Create combined batch for input to Hessian calculation
        x_in = torch.cat([
            zFull.transpose(1, 2).reshape(batch_size * time_dim, self.d),
            n_steps.repeat(batch_size, 1)
        ], dim=1).to(self.device)
    
        # x_in requires_grad
        #x_in = x_in.clone().requires_grad_(True)

        # Define function for which to compute the Hessian
        def func(batch_x_in):
            return self.forward(batch_x_in.reshape(1, self.d+1)).sum()
    
        # Compute the Hessians function for all arguments but the last
        hess = vmap(torch.func.hessian(func))(x_in)
        #hess = torch.randn(batch_size, time_dim, self.d+1, self.d+1, device=self.device)
        hess = hess.reshape(batch_size, time_dim, self.d+1, self.d+1)
    
        # Extract spatial dimensions and time steps
        hess = hess[:, :, :self.d, :self.d]
    
        # Average over time steps
        hess = hess.mean(dim=1).reshape(-1, dim, dim)

        transformed = zFull[:, :dim]

        unif = jacobian_gaussian_to_sphere(transformed[:, :, -1].clone().reshape(-1, dim)) #.reshape(x.shape[0], dim, dim)

        full_hessians = torch.einsum('bij,bjk->bik', unif, hess)
        metric_t = torch.einsum('bji,bjk->bik', full_hessians, full_hessians)

        # add a small unit matrix to ensure numerical stability

        metric_t = metric_t + 1e-6 * torch.eye(dim, device=self.device).unsqueeze(0)

        return metric_t.detach() #, hess, unif
    
    def metric_tensor(self, x):
        # flatten x and reshape into original shape
        x_dims = torch.tensor(x.shape)
        x_new_dims = torch.concat([x_dims, x_dims[-1].reshape(-1)])

        #x = x.flatten(0, 1)
        x = x.reshape(-1, x.shape[-1])

        # if x is too large, split up in batches of size 1024
        if x.shape[0] > 1024:
            num_batches = int(x.shape[0] / 1024) + 1
            batch_size = int(x.shape[0] / num_batches)
            hess = torch.zeros(x.shape[0], x.shape[1], x.shape[1], device=self.device)

            for i in range(num_batches):
                start_idx = i * batch_size
                end_idx = min((i + 1) * batch_size, x.shape[0])
                hess[start_idx:end_idx] = self.fullHessian(x[start_idx:end_idx], 8)

            hess = hess.reshape(-1, x.shape[1], x.shape[1])

        else:
            # compute the full Hessian
            hess = self.fullHessian(x, 8)

        g = hess.reshape(*x_new_dims)
        return g
    

# define C cost
def C(z):
    """Expected negative log-likelihood; see Eq.(3) in the paper"""
    d = z.shape[1] - 3
    l = z[:, d]  # log-det

    return -(torch.sum(-0.5 * math.log(2 * math.pi) - torch.pow(z[:, 0:d], 2) / 2, 1, keepdim=True) + l.unsqueeze(1))


def OTFlowProblem(x, Phi, tspan, nt, stepper="rk4", alph=[1.0, 1.0, 1.0]):
    """
    Evaluate objective function of OT Flow problem; see Eq. (8) in the paper.

    :param x:       input data tensor nex-by-d
    :param Phi:     neural network
    :param tspan:   time range to integrate over, ex. [0.0 , 1.0]
    :param nt:      number of time steps
    :param stepper: string "rk1" or "rk4" Runge-Kutta schemes
    :param alph:    list of length 3, the alpha value multipliers
    :return:
        Jc - float, objective function value dot(alph,cs)
        cs - list length 5, the five computed costs
    """
    h = (tspan[1] - tspan[0]) / nt

    # initialize "hidden" vector to propagate with all the additional dimensions for all the ODEs
    z = nn.functional.pad(x, (0, 3, 0, 0), value=0)

    tk = tspan[0]

    if stepper == 'rk4':
        for k in range(nt):
            z = stepRK4(odefun, z, Phi, alph, tk, tk + h)
            tk += h
    elif stepper == 'rk1':
        for k in range(nt):
            z = stepRK1(odefun, z, Phi, alph, tk, tk + h)
            tk += h

    # ASSUME all examples are equally weighted
    costL = torch.mean(z[:, -2])
    costC = torch.mean(C(z))
    costR = torch.mean(z[:, -1])

    cs = [costL, costC, costR]

    # return dot(cs, alph)  , cs
    return sum(i[0] * i[1] for i in zip(cs, alph)), cs


def compute_loss(net, x, nt):
    Jc, cs = OTFlowProblem(x, net, [0, 1], nt=nt, stepper="rk4", alph=net.alph)
    # detach cs and throw away
    cs = [c.detach().cpu().numpy() for c in cs]
    return Jc #, cs


def train_OTFlow(net, optimizer, X_train, X_val, epochs = 1500, nt=8, nt_val=8, drop_freq=100, lr_drop=2, batch_size=1024):
    net.train()

    loss_hist = {
        'train': [],
        'val': []
    }

    val_best = 9999999
    epoch_best = 0
    for itr in range(0, epochs):
        #batches
        loss = 0.0
        for batch_start in range(0, X_train.shape[0], batch_size):
            batch_end = min(batch_start + batch_size, X_train.shape[0])
            batch_data = X_train[batch_start:batch_end]

            optimizer.zero_grad()
            l = compute_loss(net, batch_data, nt)
            l.backward()
            optimizer.step()

            loss += l.item() * (batch_end - batch_start) / X_train.shape[0]

        if (itr % 100 == 99) or (itr == 0):
            net.eval()
            #Jc, cs = compute_loss(net, X_train, nt)
            with torch.no_grad():
                Jcval = compute_loss(net, X_val, nt_val)
                print(f"Iteration {itr} | Train Loss: {loss} | Validation Loss: {Jcval.item()}")
            if Jcval < val_best:
                val_best = Jcval
                epoch_best = 0
            else:
                epoch_best += 1

            # improve on the early stopping
            if epoch_best >= 5:
                # stop training
                print(f"Early stopping at iteration {itr}")
                # break the loop
                break
            net.train()

            loss_hist['train'].append(loss)
            loss_hist['val'].append(Jcval.item())

        if itr % drop_freq == 0:
            for param_group in optimizer.param_groups:
                param_group['lr'] = param_group['lr'] / lr_drop

    net.eval()

    return loss_hist




import torch
from scipy.special import gamma, gammainc, gammaincinv

'''
Helper functions for converting points from a gaussian distribution to a uniform distribution
on the unit sphere and vice versa.
'''

# vectorized implementation of the regularized lower incomplete gamma function
def vec_gammainc(a, x, num_steps=100):
    """
    Differentiable implementation of the regularized lower incomplete gamma function P(a, x).
    Supports vectorized inputs for `a` and `x`.

    Arguments:
        a: Shape parameter (tensor), must be positive.
        x: Upper limit of integration (tensor), must be non-negative.
        num_steps: Number of steps for numerical integration.
    Returns:
        Regularized lower incomplete gamma function P(a, x).
    """
    device = x.device

    #is a not tensor, make it
    if not torch.is_tensor(a):
        a = torch.tensor(a, dtype=torch.float32, device=device) 

    #assert torch.all(a > 0), "Parameter 'a' must be positive."
    #assert torch.all(x >= 0), "Parameter 'x' must be non-negative."

    # Prepare integration steps
    t = torch.linspace(0, 1, num_steps, device=device, dtype=x.dtype)  # Shape: (num_steps,)
    dt = x.unsqueeze(-1) * t  # Scale t to [0, x], Shape: (..., num_steps)

    # Compute integrand t^(a-1) * exp(-t)
    a_expanded = a.unsqueeze(-1)  # Shape: (..., 1)
    integrand = (dt ** (a_expanded - 1)) * torch.exp(-dt)  # Shape: (..., num_steps)
    #integrand[torch.isnan(integrand)] = 0  # Handle NaNs for edge cases (e.g., 0^(-1)).
    integrand = torch.where(torch.isnan(integrand), torch.zeros_like(integrand), integrand)  # Handle NaNs for edge cases (e.g., 0^(-1)).


    # Numerical integration using the trapezoidal rule
    trapezoid = (integrand[..., 1:] + integrand[..., :-1])  # Shape: (..., num_steps - 1)
    integral = trapezoid.sum(dim=-1) * (x / num_steps)  # Shape: (...)

    # Regularization with the gamma function
    gamma_a = torch.exp(torch.special.gammaln(a))  # Compute Gamma(a), Shape: (...)
    return integral / gamma_a  # Shape: (...)


def gaussian_to_sphere(X: torch.Tensor) -> torch.Tensor:
    # Convert a multivariate gaussian of any dimension d to a d-dimensional sphere
    d = X.shape[1]
    # Compute the norm of each row
    norm = torch.norm(X, dim=1, keepdim=True)

    # compute cdf of each point
    cdf = gammainc(d/2, norm**2/2) ** (1/d)

    # calculate the new point with the adjusted radius
    X_sphere = X / norm * cdf
    return X_sphere


def sphere_to_gaussian(X: torch.Tensor) -> torch.Tensor:
    # Convert a d-dimensional sphere to a multivariate gaussian of any dimension d
    d = X.shape[1]
    # Compute the norm of each row
    norm = torch.norm(X, dim=1, keepdim=True)

    # check if any norm is 1 and set to 0.9999 to avoid infinities
    if torch.any(norm >= 1):
        norm_new = torch.where(norm >= 1, torch.tensor([0.9999]), norm)

        X = X / norm * norm_new
        norm = norm_new

    # compute cdf of each point
    inv_cdf = (gammaincinv(d/2, (norm ** d))*2)**0.5

    # calculate the new point with the adjusted radius
    X_gaussian = X / norm * inv_cdf
    return X_gaussian

# vecotrized implementation of the jacobian of the transformation from a multivariate gaussian to a d-dimensional sphere
def jacobian_gaussian_to_sphere(X: torch.Tensor) -> torch.Tensor: #at point x
    # Compute the jacobian of the transformation from a multivariate gaussian to a d-dimensional sphere
    n = X.shape[0]
    d = X.shape[1]
    norm = torch.norm(X, dim=1, keepdim=True).to(X.device)

    gammainc_d2 = vec_gammainc(d/2, norm**2/2).to(X.device)
    gamma_d2 = torch.tensor(gamma(d/2), dtype=torch.float32, device=X.device)

    constants = 2**(2-d) / gamma_d2 * 1/d

    nex_constants_1 = norm**(d-3) * torch.exp(-norm**2/2) * gammainc_d2**(1/d - 1)
    nex_constants_2 = norm**(-3) * gammainc_d2**(1/d)

    nex_unity = norm**-1 * gammainc_d2**(1/d)

    base_matrix = torch.einsum('bi,bj->bij', X, X)
    
    term1 = torch.einsum('bi,bij->bij', constants * nex_constants_1, base_matrix)
    term2 = torch.einsum('bi,bij->bij', -nex_constants_2, base_matrix)
    term3 = torch.einsum('bi,bij->bij', nex_unity, torch.eye(d, device=X.device).repeat(n, 1, 1))

    J = term1 + term2 + term3

    return J


import torch
import numpy as np

#from typing import Union
from scipy.spatial import KDTree
from scipy.sparse.csgraph import dijkstra
from scipy.sparse import csr_matrix
#from scipy.stats import shapiro, combine_pvalues    
#from pingouin import multivariate_normality
#from tqdm import tqdm
#from LAMINAR.Flow.planarCNF import PlanarCNF, train_PlanarCNF
from LAMINAR.Flow.OTFlow import Phi, train_OTFlow, integrate
from LAMINAR.utils.gaussian2uniform import gaussian_to_sphere
from LAMINAR.utils.geodesics import geodesic_length, geodesic_path, geodesic_straight_line
from LAMINAR.utils.dijkstra import dijkstra as dijkstra_laminar

'''
Implementation of the LAM algorithm using a normalizing flow to transform the data
'''
class LAMINAR():
    def __init__(self,
                 data, 
                 alph = [1.0, 100.0, 5.0],
                 nt = 8, 
                 nt_val = 8,
                 nTh = 3,
                 m = 32,
                 lr = 0.1,
                 drop_freq = 100,
                 lr_drop = 2,
                 k_neigh = 10,
                 epochs = 1500,
                 batch_size = 1024,
                 save_distance_matrix = False):
        
        self.device = data.device
        self.data = data

        self.alph = alph
        self.nt = nt
        self.nt_val = nt_val
        self.nTh = nTh
        self.m = m
        self.lr = lr
        self.drop_freq = drop_freq
        self.lr_drop = lr_drop
        self.k_neigh = k_neigh 
        self.epochs = epochs

        self.batch_size = batch_size

        self.save_distance_matrix = save_distance_matrix

        self.d = self.data.shape[1]
        self.n = self.data.shape[0]

        # split the data into training and validation
        self.data_train = self.data[:int(self.n*0.8)]
        self.data_val = self.data[int(self.n*0.8):]

        # initialize the normalizing flow
        self.net = Phi(self.nTh, self.m, self.d, alph=self.alph, device=self.device)
        self.optimizer = torch.optim.Adam(self.net.parameters(), lr=self.lr)

        # train the model
        self.loss_hist = train_OTFlow(self.net, self.optimizer, self.data_train, self.data_val, self.epochs, self.nt, self.nt_val, self.drop_freq, self.lr_drop, self.batch_size)

        # set up the graph
        self.set_up_graph()


    def switch_device(self, device):
        self.device = device
        self.net.to(device)
        self.data = self.data.to(device)
        
    
    def set_up_graph(self):
        self.X_pushed = integrate(self.data, self.net, [0, 1], nt=self.nt, stepper="rk4", alph=self.alph, intermediates=False).cpu().detach()[:, :self.d]
        self.X_pushed = gaussian_to_sphere(self.X_pushed)

        kdt = KDTree(self.X_pushed)
        _, neighs = kdt.query(self.X_pushed, k=self.k_neigh)

        # get all start points
        starts = self.data[neighs[:, 0]]
        # repeat every point in starts for k_neigh times so that the same point appears k_neigh times right after each other
        starts = starts.repeat_interleave(self.k_neigh, dim=0)
        ends = self.data[neighs.flatten()]

        distances = geodesic_straight_line(starts, ends, self.net.metric_tensor, inbetween=1).flatten()
        row_indices = torch.arange(neighs.shape[0]).repeat_interleave(neighs.shape[1])
        col_indices = neighs.flatten()

        self.graph = csr_matrix((distances.cpu().numpy().astype(np.float32), (row_indices, col_indices)), shape=(self.n, self.n))

        # symmetrize 
        self.graph = self.graph.maximum(self.graph.transpose()).tocsc()

        if self.save_distance_matrix:
            self.dist_matrix, self.predecessors = dijkstra(self.graph, return_predecessors=True)
            self.dist_matrix = self.dist_matrix.astype(np.float32)
            self.predecessors = self.predecessors.astype(np.int32)

    
    def expand_graph(self, additional_points):
        # additional_points is an array of shape (m, d) of points which temporarily need to be added to the graph
        # returns the expanded graph, the distance matrix and the predecessors

        #additional_points = additional_points.reshape(-1, self.d)
        expanded_data = torch.cat([self.data, additional_points], dim=0)
        additional_points_pushed = integrate(additional_points, self.net, [0, 1], nt=self.nt_val, stepper="rk4", alph=self.alph, intermediates=False).cpu().detach()[:, :self.d]
        additional_points_pushed = gaussian_to_sphere(additional_points_pushed)

        #expanded_data_pushed = torch.concatenate([self.X_pushed, additional_points_pushed], dim=0)

        #kdt = KDTree(expanded_data_pushed)
        kdt = KDTree(self.X_pushed)
        _, neighs = kdt.query(additional_points_pushed, k=self.k_neigh)

        starts = additional_points.repeat_interleave(self.k_neigh, dim=0)
        ends = expanded_data[neighs.flatten()]

        distances = geodesic_straight_line(starts, ends, self.net.metric_tensor, inbetween=1)
        row_indices = torch.arange(neighs.shape[0]).repeat_interleave(neighs.shape[1])
        col_indices = neighs.flatten()

        # expanded graph is self.graph with additional elements

        expanded_graph = csr_matrix((distances.cpu().numpy().astype(np.float32), (row_indices + self.n, col_indices)), shape=(expanded_data.shape[0], expanded_data.shape[0]))
        expanded_graph = expanded_graph.tolil().astype(np.float32)

        # print info of expanded graph
        #print("Expanded graph shape: ", expanded_graph.shape)
        #print("Expanded graph type: ", expanded_graph.dtype)
        #
        # same for graph
        #print("Graph shape: ", self.graph.shape)
        #print("Graph type: ", self.graph.dtype)


        #expanded_graph[:self.n, :self.n] = self.graph
        self.graph = self.graph.tolil().astype(np.float32)

        block_size = 1024  # Adjust block size based on available memory
        for i in range(0, self.n, block_size):
            for j in range(0, self.n, block_size):
                # Calculate the actual block size for the current slice
                i_end = min(i + block_size, self.n)
                j_end = min(j + block_size, self.n)

                # Print shapes for debugging
                #print(f"Assigning block: [{i}:{i_end}, {j}:{j_end}]")
                #print(f"Source shape: {self.graph[i:i_end, j:j_end].shape}")
                #print(f"Target shape: {expanded_graph[i:i_end, j:j_end].shape}")

                # Assign the block
                expanded_graph[i:i_end, j:j_end] = self.graph[i:i_end, j:j_end]

        expanded_graph = expanded_graph.tocsr()
        expanded_graph = expanded_graph.maximum(expanded_graph.transpose()).tocsc()

        self.graph = self.graph.tocsr()

        if self.save_distance_matrix:
            dist_matrix, predecessors = dijkstra(expanded_graph, return_predecessors=True)
            dist_matrix = dist_matrix.astype(np.float32)
            predecessors = predecessors.astype(np.int32)

            return expanded_graph, dist_matrix, predecessors 
        
        else:
            return expanded_graph, None, None


    def check_expansion(self, points):
        # for each point, check wether it is in the data, if so note the indices
        # else add the point to the graph

        # check if the points are in the data
        kdt = KDTree(self.data)
        dists, idx = kdt.query(points, k=1)

        dists = torch.tensor(dists)

        # args where dists is zero
        in_data = torch.where(dists == 0)[0]
        at_indices = idx[in_data]
        at_indices = torch.tensor(at_indices).reshape(-1)

        not_in_data = torch.where(dists != 0)[0]

        # expand graph by not_in_data
        # if not_in_data is not empty
        if not_in_data.shape[0] != 0:
            # expand the graph
            expanded_graph, dist_matrix, predecessors = self.expand_graph(points[not_in_data])

            # note the indices
            not_in_data_idx = torch.arange(self.n, self.n + not_in_data.shape[0]) #[not_in_data]

            # concat the windices of points in the data
            idx = torch.cat([at_indices, not_in_data_idx])

            if self.save_distance_matrix:
                # make the distance matrix a tensor and return
                dist_matrix = torch.tensor(dist_matrix)

                return expanded_graph, idx.tolist(), dist_matrix, predecessors
            
            else:
                return expanded_graph, idx.tolist(), None, None
        
        else:
            # no extension needed, just return the indices and the distance matrix
            idx = at_indices

            if self.save_distance_matrix:
                dist_matrix = torch.tensor(self.dist_matrix)

                return self.graph, idx.tolist(), dist_matrix, self.predecessors
            
            else:
                return self.graph, idx.tolist(), None, None

    
    def query(self, start, k=None):     # TODO expand to add point besides the data
        # start is an array of shape (m, d)
        # if shape is (d,) reshape to (1, d)
        if start.shape == (self.d,):
            start = start.reshape(1, self.d)

        # calculate the k nearest points and their distance for each start point
        # expand the graph with the new points
        expanded_graph, idx_points, dist_matrix, _ = self.check_expansion(start)

        if self.save_distance_matrix:
            dists = dist_matrix[idx_points].detach() # shape (m, n)
        
        else:
            dists, _ = dijkstra_laminar(expanded_graph, idx_points, [i for i in range(self.n)])
            dists = torch.tensor(np.array(dists)).reshape(-1, self.n)

        dists, idx = torch.sort(dists, dim=1)
        if k is not None:
            return idx[:, :k], dists[:, :k]

        else:
            return idx, dists

    
    def distance_approx(self, start, end, return_path=False):
        # expand graph by end and start points
        start = start.unsqueeze(0) if start.dim() == 1 else start
        end = end.unsqueeze(0) if end.dim() == 1 else end
        all_points = torch.cat([self.data, start, end], dim=0)
        
        #_, dist_matrix, predecessors = self.expand_graph(torch.cat([start, end], dim=0))

        expanded_graph, idx, dist_matrix, predecessors = self.check_expansion(torch.cat([start, end], dim=0))

        start_idx = idx[-2]
        end_idx = idx[-1]    
    
        if self.save_distance_matrix:
            if return_path:
                path_idx = [end_idx]
                current = end_idx

                while current != start_idx:
                    current = predecessors[start_idx, current]
                    path_idx.append(current)

                path_idx = torch.tensor(path_idx).flip(0)
                path = all_points[path_idx]

                #dist = geodesic_length(path.reshape(1, path.shape[0], self.d), start, end, self.net.metric_tensor)
                dist = dist_matrix[start_idx, end_idx]

                return dist, path

            else:
                dist = dist_matrix[start_idx, end_idx]
                return dist
            
        else:
            dists, path_idx = dijkstra_laminar(expanded_graph, [start_idx], [end_idx])
            if return_path:
                paths = all_points[path_idx[0]]
                return np.array(dists[0]), paths
            else:   
                return np.array(dists[0])


    def distance_smooth(self, start, end):
        _, path = self.distance_approx(start, end, return_path=True)
        points, _ = geodesic_path(start, end, self.net.metric_tensor, lr=1e-2, initial_guess=path, max_iter=1000)
        dist = geodesic_length(points[1:-1].reshape(1, points[1:-1].shape[0], self.d), start, end, self.net.metric_tensor)

        return dist[0].detach(), points.detach()
    


import numpy as np
import numba as nb


def dijkstra(graph, start_nodes, end_nodes):
    n = graph.shape[0]
    indptr = graph.indptr.astype(np.int32)
    indices = graph.indices.astype(np.int32)
    data = graph.data.astype(np.float32)
    return _dijkstra(n, indptr, indices, data, start_nodes, end_nodes)

@nb.njit
def _dijkstra(n, indptr, indices, data, start_nodes, end_nodes):
    inf = float('inf')

    distance_array = []
    path_array = []

    for start in start_nodes:
        distances = np.full(n, inf, dtype=np.float32)
        predecessors = np.full(n, -1, dtype=np.int32)
        visited = np.zeros(n, dtype=np.bool_)

        distances[start] = 0.0

        for _ in range(n):
            # Stop early, if all end nodes have been visited
            all_end_nodes_visited = True
            for end in end_nodes:
                if not visited[end]:
                    all_end_nodes_visited = False
                    break
                
            if all_end_nodes_visited:
                break

            # Find the node with the smallest distance that hasn't been visited
            min_dist = inf
            u = -1
            for i in range(n):
                if not visited[i] and distances[i] < min_dist:
                    min_dist = distances[i]
                    u = i

            if u == -1 or min_dist == inf:
                break

            visited[u] = True

            # Update the distances to the neighbors
            for idx in range(indptr[u], indptr[u + 1]):
                v = indices[idx]
                length = data[idx]
                if not visited[v] and distances[u] + length < distances[v]:
                    distances[v] = distances[u] + length
                    predecessors[v] = u

        # Extract the shortest path distances to the end nodes
        end_distances = np.array([distances[end] for end in end_nodes])

        end_paths = []
        for end in end_nodes:
            path = []
            while end != -1:
                path.append(end)
                end = predecessors[end]
            path.reverse()
            end_paths.append(path)

        distance_array.append(end_distances)
        path_array.append(end_paths)

    return distance_array, path_array



import numpy as np
import matplotlib.pyplot as plt

def hsv_to_rgb(H, S, V):
    h = np.floor(H/60)
    f = H/60 - h
    p = V * (1 - S)
    q = V * (1 - f * S)
    t = V * (1 - (1 - f) * S)

    if (h == 0) or (h == 6):
        return V, t, p
    if h == 1:
        return q, V, p
    if h == 2:
        return p, V, t
    if h == 3:
        return p, q, V
    if h == 4:
        return t, p, V
    if h == 5:
        return V, p, q

# plot each colour on a circle where the angle corresponds to the hue
def plot_reference():
    fig, ax = plt.subplots(1, 1, figsize=(5, 5))

    for r in np.linspace(0, 1, 20):
        S = r
        for angle in range(360):
            H = angle * 2 % 360
            R, G, B = hsv_to_rgb(H, S, 1)
            ax.plot(r*np.cos(angle/180*np.pi), r*np.sin(angle/180*np.pi), marker='o', color=(R, G, B), markersize=10)
    # axis off
    ax.axis('off')
    plt.show()

# only two dim
def get_cov_colours(
            data,               # original data in original space
            metric_tensor             
            ):
    colours = []

    for point in range(len(data)):
        eigenvalues, eigenvectors = np.linalg.eig(metric_tensor[point])

        # volume of the ellipse
        volume = np.sqrt(np.prod(eigenvalues))

        # scale it so a large value goes asymptotically to 1
        S = 1 - np.exp(-volume)
        
        eig_idx = np.argmax(eigenvalues)
        vec = eigenvectors[:, eig_idx]

        angle = np.arctan2(vec[1], vec[0])
        angle = np.degrees(angle)

        if angle < 0:
            angle += 360

        colour_angle = (angle % 180) * 2

        H = colour_angle

        #print(eigenvalues)

        ratio = eigenvalues.max() / (eigenvalues.min())
        #print(ratio)

        # ratio of 1 means black i.e. V = 0
        # ratio of inf means white i.e. V = 1
        V = 1 - (1/ratio)**0.25
        #print(H, S)
        R, G, B = hsv_to_rgb(H, V, 1)
        colours.append((R, G, B))
    return colours